from typing import Any, Dict, List, Optional
from axe_selenium_python.axe import Axe
from pydantic.main import BaseModel
from pydantic import Field
from selenium.webdriver.remote.webdriver import WebDriver


class AxeRelatedNode(BaseModel):
    html: str
    target: List[str]


class AxeSubNode(BaseModel):
    data: Optional[Any]
    id: str
    impact: Optional[str]
    message: str
    related_nodes: List[AxeRelatedNode] = Field(alias='relatedNodes')


class AxeNode(BaseModel):
    all: List[AxeSubNode]
    any: List[AxeSubNode]
    html: str
    impact: Optional[str]
    none: List[AxeSubNode]
    target: List[str]


class AxeAudit(BaseModel):
    description: str
    help: str
    help_url: str = Field(alias='helpUrl')
    id: str
    impact: Optional[str]
    nodes: List
    tags: List[str]


class AxeNodeViolation(AxeNode):
    failure_summary: str = Field(alias='failureSummary')


class AxeAuditViolation(AxeAudit):
    nodes: List[AxeNodeViolation]


class AxeReport(BaseModel):
    """ The aXe Audit Report in a user-friendly object. """
    inapplicable: List[AxeAudit]
    incomplete: List[AxeAudit]
    passes: List[AxeAudit]
    violations: List[AxeAuditViolation]
    timestamp: str
    url: str


class PyleniumAxe:
    """ The Pylenium abstraction of the axe-selenium-python package. """

    def __init__(self, webdriver: WebDriver):
        self.webdriver = webdriver

    def run(self, name: Optional[str] = None, context: Optional[Dict] = None, options: Optional[Dict] = None) -> AxeReport:
        """ Run the aXe audit and return an AxeReport object with the results.

        For more info on the `context` and `options` parameters, visit the aXe official docs:
        https://github.com/dequelabs/axe-core/blob/master/doc/API.md#parameters-axerun

        Args:
            name: The file path and name of the report to save as a JSON. If not provided, the file is not saved.
            context: The dictionary of page part(s), by CSS Selectors, to include or exclude in the audit.
            options: The dictionary of aXe options to include in the audit.

        Examples:
            # Save the report to export or share
            PyleniumAxe(py.webriver).run(name='a11y_report.json')

            # Use the AxeReport for a hard assertion
            report = PyleniumAxe(py.webdriver).run(name='ally_report.json')
            violation_count = len(report.violations)
            assert violation_count == 0, f'{violation_count} violation(s) found!'

        Raises:
            FileNotFoundError if the given `name` is within a directory that doesn't exist.
                * Include the file extension (e.g., axe_audit.json)
                * If the file does not exist, it will be created
                * However, if a directory does not exist, the file is not created and the above Error is raised
        """
        axe = Axe(self.webdriver)
        axe.inject()
        results = axe.run(context=context, options=options)
        if name:
            axe.write_results(results, name)
        return AxeReport(**results)



from selenium.webdriver.remote.webdriver import WebDriver, By
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support.wait import WebDriverWait
from selenium.common.exceptions import StaleElementReferenceException
from pylenium import utils


def inject(driver: WebDriver, version="3.5.1", timeout=10):
    """Inject the given jQuery version to the current context and any iframes within it.

    Args:
        driver: The instance of WebDriver to attach to.
        version: The jQuery version. (Default is 3.5.1)
        timeout: The max number of seconds to wait for jQuery to be loaded. (Default is 10)
    """
    jquery_url = f"https://code.jquery.com/jquery-{version}.min.js"
    load_jquery = utils.read_script_from_file("load_jquery.js")
    driver.execute_async_script(load_jquery, jquery_url, None)
    WebDriverWait(driver, timeout).until(
        lambda drvr: drvr.execute_script('return typeof(jQuery) !== "undefined";'),
        message='jQuery was "undefined" which means it did not load within timeout.',
    )
    iframes = driver.find_elements(By.TAG_NAME, "iframe")
    for iframe in iframes:
        try:
            driver.execute_async_script(load_jquery, jquery_url, iframe)
        except StaleElementReferenceException:
            pass


def exists(driver: WebDriver) -> str:
    """Checks if jQuery exists in the current context.

    Returns:
        The version if found, else returns an empty string
    """
    version = driver.execute_script("return jQuery().jquery;")
    return version if version is not None else ""


def drag_and_drop(driver: WebDriver, drag_element: WebElement, drop_element: WebElement, version="3.5.1", timeout=10):
    """Simulate Drag and Drop using jQuery.

    Args:
        driver: The driver that will simulate the drag and drop.
        drag_element: The element to be dragged.
        drop_element: The element to drop to.
        version: The jQuery version to use.
        timeout: The max of number of seconds to wait for jQuery to be loaded.
    """
    inject(driver, version)
    dnd_js = utils.read_script_from_file("drag_and_drop.js")
    driver.execute_script(
        dnd_js + "jQuery(arguments[0]).simulateDragDrop({ dropTarget: arguments[1] });", drag_element, drop_element
    )



import pytest
from pylenium.driver import Pylenium


@pytest.fixture
def sauce(py: Pylenium) -> Pylenium:
    """Login to saucedemo.com as standard user."""
    py.visit("https://www.saucedemo.com/")
    py.get("#user-name").type("standard_user")
    py.get("#password").type("secret_sauce")
    py.get("#login-button").click()
    yield py
    py.get("#react-burger-menu-btn").click()
    py.get("#logout_sidebar_link").should().be_visible().click()


def test_add_to_cart_css(sauce: Pylenium):
    """Add an item to the cart. The number badge on the cart icon should increment as expected."""
    sauce.get("[id*='add-to-cart']").click()
    assert sauce.get("a.shopping_cart_link").should().have_text("1")


def test_add_to_cart_xpath(sauce: Pylenium):
    """Add 6 different items to the cart. There should be 6 items in the cart."""
    for button in sauce.findx("//*[contains(@id, 'add-to-cart')]"):
        button.click()
    sauce.getx("//a[@class='shopping_cart_link']").click()
    assert sauce.findx("//*[@class='cart_item']").should().have_length(6)



""" Chrome DevTools Protocol (CDP) introduced in Selenium 4.

Resources:
    - https://www.selenium.dev/documentation/webdriver/bidirectional/chrome_devtools/
    - https://chromedevtools.github.io/devtools-protocol/

* Currently only supports the Chrome Browser, although some chromium browsers may work as well.
"""

from typing import Dict


class CDP:
    """Chrome DevTools Protocol."""

    def __init__(self, webdriver):
        self._webdriver = webdriver

    def execute_command(self, cmd: str, cmd_args: Dict) -> Dict:
        """Execute Chrome Devtools Protocol command and get returned result.

        The command and command args should follow chrome devtools protocol domains/commands, refer to link
        https://chromedevtools.github.io/devtools-protocol/

        Args:
            cmd: The command name
            cmd_args: The command args. Pass an empty dict {} if there is no command args

        Examples:
            py.cdp.execute_command('Network.getResponseBody', {'requestId': requestId})

        Returns:
            A dict of results or an empty dict {} if there is no result to return.
            For example, to getResponseBody:
            {'base64Encoded': False, 'body': 'response body string'}
        """
        return self._webdriver.execute_cdp_cmd(cmd, cmd_args)

    def get_performance_metrics(self) -> Dict:
        """Get performance metrics from Chrome DevTools - similar to the Performance tab in Chrome.

        Examples:
            metrics = py.cdp.get_performance_metrics()

        Returns:
            A dict of performance metrics including 'ScriptDuration', 'ThreadTime', 'ProcessTime', and 'DomContentLoaded'.

            {'metrics': [
                {'name': 'Timestamp', 'value': 425608.80694},
                {'name': 'AudioHandlers', 'value': 0},
                {'name': 'ThreadTime', 'value': 0.002074},
                ...
                ]
            }
        """
        # The commented out code below should have been executed prior to this function call.
        # self._webdriver.execute_cdp_cmd("Performance.enable", {})
        return self._webdriver.execute_cdp_cmd("Performance.getMetrics", {})



""" Factory to build WebDrivers. """
from typing import List, Optional

from selenium import webdriver
from selenium.webdriver import Chrome
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.chrome.service import Service as ChromeService
from selenium.webdriver.edge.service import Service as EdgeService
from selenium.webdriver.edge.options import Options as EdgeOptions
from webdriver_manager.chrome import ChromeDriverManager
from webdriver_manager.firefox import GeckoDriverManager
from webdriver_manager.microsoft import IEDriverManager, EdgeChromiumDriverManager
from webdriver_manager.opera import OperaDriverManager
from pylenium.config import PyleniumConfig


class Browser:
    """ENUM of supported browsers."""

    CHROME = "chrome"
    EDGE = "edge"
    FIREFOX = "firefox"
    IE = "ie"
    OPERA = "opera"


def build_capabilities(browser, capabilities: dict):
    """Build the capabilities dictionary for WebDriver.

    Args:
        browser: The name of the browser.
        capabilities: The dict of capabilities to include.

    Examples:
        caps = WebDriverFactory().build_capabilities({'enableVNC': True, 'enableVideo': False})
    """
    caps = {}

    if browser == Browser.CHROME:
        caps.update(webdriver.DesiredCapabilities.CHROME.copy())
    elif browser == Browser.FIREFOX:
        caps.update(webdriver.DesiredCapabilities.FIREFOX.copy())
    elif browser == Browser.IE:
        caps.update(webdriver.DesiredCapabilities.INTERNETEXPLORER.copy())
    elif browser == Browser.OPERA:
        caps.update(webdriver.DesiredCapabilities.OPERA.copy())
    elif browser == Browser.EDGE:
        caps.update(webdriver.DesiredCapabilities.EDGE.copy())
    else:
        raise ValueError(f"{browser} is not supported.")

    if capabilities:
        caps.update(capabilities)

    return caps


def build_options(
    browser,
    browser_options: List[str],
    experimental_options: Optional[List[dict]],
    extension_paths: Optional[List[str]],
):
    """Build the Options object for Chrome or Firefox.

    Args:
        browser: The name of the browser.
        browser_options: The list of options/arguments to include.
        experimental_options: The list of experimental options to include.
        extension_paths: The list of extensions to add to the browser session.

    Examples:
        WebDriverFactory().build_options('chrome', ['headless', 'incognito'], [{'useAutomationExtension', False}])
    """
    browser = browser.lower()
    if browser == Browser.CHROME:
        options = webdriver.ChromeOptions()
    elif browser == Browser.FIREFOX:
        options = webdriver.FirefoxOptions()
    elif browser == Browser.IE:
        options = webdriver.IeOptions()
    elif browser == Browser.OPERA:
        options = webdriver.ChromeOptions()
    elif browser == Browser.EDGE:
        options = EdgeOptions()
    else:
        raise ValueError(f"{browser} is not supported.")

    for option in browser_options:
        if option.startswith("--"):
            options.add_argument(option)
        else:
            options.add_argument(f"--{option}")

    if experimental_options:
        for exp_option in experimental_options:
            ((name, value),) = exp_option.items()
            options.add_experimental_option(name, value)

    if extension_paths:
        for path in extension_paths:
            options.add_extension(path)

    return options


def build_from_config(config: PyleniumConfig) -> WebDriver:
    """Build a WebDriver using PyleniumConfig.

    PyleniumConfig is built using pylenium.json and CLI args.
    """
    if config.driver.remote_url:
        return build_remote(
            browser=config.driver.browser,
            remote_url=config.driver.remote_url,
            browser_options=config.driver.options,
            capabilities=config.driver.capabilities,
            experimental_options=config.driver.experimental_options,
            extension_paths=config.driver.extension_paths,
            webdriver_kwargs=config.driver.webdriver_kwargs,
        )
    browser = config.driver.browser.lower()
    if browser == Browser.CHROME:
        return build_chrome(
            config.driver.version,
            config.driver.options,
            config.driver.experimental_options,
            config.driver.extension_paths,
            config.driver.local_path,
            config.driver.webdriver_kwargs,
        )
    elif browser == Browser.FIREFOX:
        return build_firefox(
            config.driver.version,
            config.driver.options,
            config.driver.capabilities,
            config.driver.experimental_options,
            config.driver.extension_paths,
            config.driver.local_path,
            config.driver.webdriver_kwargs,
        )
    elif browser == Browser.IE:
        return build_ie(
            config.driver.version,
            config.driver.options,
            config.driver.capabilities,
            config.driver.experimental_options,
            config.driver.extension_paths,
            config.driver.local_path,
            config.driver.webdriver_kwargs,
        )
    elif browser == Browser.OPERA:
        return build_opera(
            config.driver.version,
            config.driver.options,
            config.driver.experimental_options,
            config.driver.extension_paths,
            config.driver.local_path,
            config.driver.webdriver_kwargs,
        )
    elif browser == Browser.EDGE:
        return build_edge(
            config.driver.version,
            config.driver.options,
            config.driver.capabilities,
            config.driver.experimental_options,
            config.driver.extension_paths,
            config.driver.local_path,
            config.driver.webdriver_kwargs,
        )
    else:
        raise ValueError(f"{config.driver.browser} is not supported.")


def build_chrome(
    version: str,
    browser_options: List[str],
    experimental_options: Optional[List[dict]],
    extension_paths: Optional[List[str]],
    local_path: Optional[str],
    webdriver_kwargs: Optional[dict],
) -> Chrome:
    """Build a ChromeDriver.

    Args:
        version: The desired version of Chrome.
        browser_options: The list of options/arguments to include.
        experimental_options: The list of experimental options to include.
        extension_paths: The list of extensions to add to the browser.
        local_path: The path to the driver binary.
        webdriver_kwargs: additional keyword arguments to pass.

    Examples:
        driver = WebDriverFactory().build_chrome('latest', ['headless', 'incognito'])
    """
    options = build_options(Browser.CHROME, browser_options, experimental_options, extension_paths)
    if local_path:
        driver = webdriver.Chrome(service=ChromeService(local_path), options=options, **(webdriver_kwargs or {}))
    else:
        driver = webdriver.Chrome(
            service=ChromeService(ChromeDriverManager(version=version).install()),
            options=options,
            **(webdriver_kwargs or {}),
        )

    # enable Performance Metrics from Chrome Dev Tools
    driver.execute_cdp_cmd("Performance.enable", {})

    return driver


def build_edge(
    version: str,
    browser_options: List[str],
    capabilities: dict,
    experimental_options: Optional[List[dict]],
    extension_paths: Optional[List[str]],
    local_path: Optional[str],
    webdriver_kwargs: Optional[dict],
) -> WebDriver:
    """Build a Edge Driver.

    Args:
        version: The desired version of Edge.
        browser_options: The list of options/arguments to include.
        capabilities: The dict of capabilities to include.
        experimental_options: The list of experimental options to include.
        extension_paths: The list of extensions to add to the browser.
        local_path: The path to the driver binary.
        webdriver_kwargs: additional keyword arguments to pass.

    Examples:
        driver = WebDriverFactory().build_edge('latest', ['headless', 'incognito'], None)
    """
    caps = build_capabilities(Browser.EDGE, capabilities)
    options = build_options(Browser.EDGE, browser_options, experimental_options, extension_paths)
    if local_path:
        return webdriver.Edge(
            service=EdgeService(local_path), capabilities=caps, options=options, **(webdriver_kwargs or {})
        )
    return webdriver.Edge(
        service=EdgeService(EdgeChromiumDriverManager(version=version).install()),
        capabilities=caps,
        options=options,
        **(webdriver_kwargs or {}),
    )


def build_firefox(
    version: str,
    browser_options: List[str],
    capabilities: dict,
    experimental_options: Optional[List[dict]],
    extension_paths: Optional[List[str]],
    local_path: Optional[str],
    webdriver_kwargs: Optional[dict],
) -> WebDriver:
    """Build a FirefoxDriver.

    Args:
        version: The desired version of Firefox.
        browser_options: The list of options/arguments to include.
        capabilities: The dict of capabilities to include.
        experimental_options: The list of experimental options to include.
        extension_paths: The list of extensions to add to the browser.
        local_path: The path to the driver binary.
        webdriver_kwargs: additional keyword arguments to pass.

    Examples:
        driver = WebDriverFactory().build_firefox('latest', ['headless', 'incognito'], None)
    """
    caps = build_capabilities(Browser.FIREFOX, capabilities)
    options = build_options(Browser.FIREFOX, browser_options, experimental_options, extension_paths)
    if local_path:
        return webdriver.Firefox(
            executable_path=local_path, options=options, capabilities=caps, **(webdriver_kwargs or {})
        )
    return webdriver.Firefox(GeckoDriverManager(version=version).install(), options=options, **(webdriver_kwargs or {}))


def build_ie(
    version: str,
    browser_options: List[str],
    capabilities: dict,
    experimental_options: Optional[List[dict]],
    extension_paths: Optional[List[str]],
    local_path: Optional[str],
    webdriver_kwargs: Optional[dict],
) -> WebDriver:
    """Build an IEDriver.

    Args:
        version: The desired version of IE.
        browser_options: The list of options/arguments to include.
        capabilities: The dict of capabilities.
        experimental_options: The list of experimental options to include.
        extension_paths: The list of extensions to add to the browser.
        local_path: The path to the driver binary.
        webdriver_kwargs: additional keyword arguments to pass.

    Examples:
        driver = WebDriverFactory().build_ie('latest', ['headless'], None)
    """
    caps = build_capabilities(Browser.IE, capabilities)
    options = build_options(Browser.IE, browser_options, experimental_options, extension_paths)
    if local_path:
        return webdriver.Ie(executable_path=local_path, options=options, capabilities=caps, **(webdriver_kwargs or {}))
    return webdriver.Ie(
        IEDriverManager(version=version).install(),
        options=options,
        capabilities=caps,
        **(webdriver_kwargs or {}),
    )


def build_opera(
    version: str,
    browser_options: List[str],
    experimental_options: Optional[List[dict]],
    extension_paths: Optional[List[str]],
    local_path: Optional[str],
    webdriver_kwargs: Optional[dict],
) -> WebDriver:
    """Build an OperaDriver.

    Args:
        version: The desired version of Opera.
        browser_options: The list of options/arguments to include.
        experimental_options: The list of experimental options to include.
        extension_paths: The list of extensions to add to the browser.
        local_path: The path to the driver binary.
        webdriver_kwargs: additional keyword arguments to pass.

    Examples:
        driver = WebDriverFactory().build_opera('latest', ['--start-maximized'], None)
    """
    options = build_options(Browser.OPERA, browser_options, experimental_options, extension_paths)
    if local_path:
        return webdriver.Opera(executable_path=local_path, options=options, **(webdriver_kwargs or {}))
    return webdriver.Opera(OperaDriverManager(version=version).install(), options=options, **(webdriver_kwargs or {}))


def build_remote(
    browser: str,
    remote_url: str,
    browser_options: List[str],
    capabilities: dict,
    experimental_options: Optional[List[dict]],
    extension_paths: Optional[List[str]],
    webdriver_kwargs: Optional[dict],
) -> WebDriver:
    """Build a RemoteDriver connected to a Grid.

    Args:
        browser: Name of the browser to connect to.
        remote_url: The URL to connect to the Grid.
        browser_options: The list of options/arguments to include.
        capabilities: The dict of capabilities to include.
        experimental_options: The list of experimental options to include.
        extension_paths: The list of extensions to add to the browser.
        webdriver_kwargs: additional keyword arguments to pass.

    Returns:
        The instance of WebDriver once the connection is successful
    """
    browser = browser.lower()
    caps = build_capabilities(browser, capabilities)
    options = build_options(browser, browser_options, experimental_options, extension_paths)

    return webdriver.Remote(
        command_executor=remote_url, desired_capabilities=caps, options=options, **(webdriver_kwargs or {})
    )



import logging
import time
from typing import Union, List, Optional

from pydantic import BaseModel, Field
from selenium.common.exceptions import TimeoutException
from selenium.webdriver.support.wait import WebDriverWait


def stopwatch(func):
    """ Stopwatch Decorator.

    Use this decorator on any function to measure how long it took
    for the function to complete. This is in seconds, but may have fractions of a second
    if the system clock provides more precision.

    Notes:
        This is _logged_, not printed to the Terminal

    Examples:
        1. How long does it take to add an item to the cart?
        @stopwatch
        def add_item_to_cart(py):
            py.get('#add-item').click()
            py.get('#added-notification').should().be_visible()

        2. How long does it take to edit an item's available stock via the API
           and see it change in the UI?
        @stopwatch
        def update_available_stock(py, item, quantity):
            payload = {'item': item, 'qty': quantity}
            api.items.update(payload)
            py.get(f'#available-stock-{item}').should().have_text(quantity)
    """
    def wrapper(*args, **kwargs):
        log = logging.getLogger('driver')
        start_time = time.time()
        func(*args, **kwargs)
        stop_time = time.time()
        func_name = func.__name__
        log.info(f'STOPWATCH - {func_name} took {stop_time - start_time} seconds')
    return wrapper


class Performance:
    """ Pylenium's Performance API. """
    def __init__(self, webdriver):
        self._webdriver = webdriver

    def _wait(self, timeout=10):
        return WebDriverWait(self._webdriver, timeout=timeout)

    def get(self):
        """ The main method used to generate a WebPerformance object from the current web page.

        Notes:
            Calling this method too soon may yield NoneTypes because the browser hasn't generated them yet.

        Examples:
            # Store the entire WebPerformance object and log it
            perf = py.performance.get()
            py.log.info(perf.dict())

            # Get a single data point from WebPerformance
            tti = py.performance.get().time_to_interactive()
        """
        return WebPerformance(
            time_origin=self.get_time_origin(),
            navigation_timing=self.get_navigation_timing(),
            paint_timing=self.get_paint_timing(),
            resources=self.get_resources()
        )

    def get_time_origin(self) -> float:
        """ Returns the timeOrigin precision value.

        This is the high resolution timestamp of the start time of the performance measurement.
        """
        js = 'return window.performance.timeOrigin;'
        time_origin = self._wait().until(lambda driver: driver.execute_script(js), 'Time Origin not generated yet')
        return time_origin

    def get_navigation_timing(self):
        """ Return the PerformanceNavigationTiming object as a Python object. """
        js = 'return window.performance.getEntriesByType("navigation")[0];'
        navigation = self._wait().until(lambda driver: driver.execute_script(js), 'NavigationTiming not generated yet')
        return NavigationTiming(**navigation)

    def get_paint_timing(self):
        """ Return the PerformancePaintTiming object as a Python object. """
        js = 'return window.performance.getEntriesByName("first-contentful-paint")[0];'
        paint = self._wait().until(lambda driver: driver.execute_script(js), 'PaintTiming not generated yet')
        return PaintTiming(**paint)

    def get_resources(self):
        """ Return a list of PerformanceResourceTiming objects as Python objects. """
        js = 'return window.performance.getEntriesByType("resource");'
        try:
            resources = self._wait().until(
                lambda driver: driver.execute_script(js),
                message='Resources not generated yet or there are none')
            return [ResourceTiming(**resource) for resource in resources]
        except TimeoutException:
            return None  # because there were no Resources captured for the current web page


class NavigationTiming(BaseModel):
    """ The PerformanceNavigationTiming Representation.

    Metrics regarding the browser's document navigation events

    References:
        https://developer.mozilla.org/en-US/docs/Web/API/PerformanceNavigationTiming
    """
    connect_end: float = Field(alias='connectEnd')
    connect_start: float = Field(alias='connectStart')
    decoded_body_size: Union[int, float] = Field(alias='decodedBodySize')
    dom_complete: float = Field(alias='domComplete')
    dom_content_loaded_event_end: float = Field(alias='domContentLoadedEventEnd')
    dom_content_loaded_event_start: float = Field(alias='domContentLoadedEventStart')
    time_to_interactive: float = Field(alias='domInteractive')
    domain_lookup_end: float = Field(alias='domainLookupEnd')
    domain_lookup_start: float = Field(alias='domainLookupStart')
    duration: float
    encoded_body_size: Union[int, float] = Field(alias='encodedBodySize')
    entry_type: str = Field(alias='entryType')
    fetch_start: float = Field(alias='fetchStart')
    initiator_type: str = Field(alias='initiatorType')
    load_event_end: float = Field(alias='loadEventEnd')
    load_event_start: float = Field(alias='loadEventStart')
    name: str
    next_hop_protocol: str = Field(alias='nextHopProtocol')
    redirect_count: int = Field(alias='redirectCount')
    redirect_end: int = Field(alias='redirectEnd')
    redirect_start: int = Field(alias='redirectStart')
    request_start: float = Field(alias='requestStart')
    response_end: float = Field(alias='responseEnd')
    response_start: float = Field(alias='responseStart')
    secure_connection_start: float = Field(alias='secureConnectionStart')
    server_timing: List = Field(alias='serverTiming')
    start_time: int = Field(alias='startTime')
    transfer_size: Union[int, float] = Field(alias='transferSize')
    type: str
    unload_event_end: int = Field(alias='unloadEventEnd')
    unload_event_start: int = Field(alias='unloadEventStart')
    worker_start: Union[int, float] = Field(alias='workerStart')


class PaintTiming(BaseModel):
    """ The PerformancePaintTiming Representation.

    Provides timing information about "paint" (also called "render") operations during web page construction.

    References:
        https://developer.mozilla.org/en-US/docs/Web/API/PerformancePaintTiming
    """
    duration: float
    entry_type: str = Field(alias='entryType', default='paint')
    name: str = Field(default='first-contentful-paint')
    start_time: float = Field(alias='startTime')


class ResourceTiming(BaseModel):
    """ The PerformanceResourceTiming Representation.

    Detailed network timing data regarding the loading of an application's resources.
    An application can use the timing metrics to determine, for example, the length of time it takes
    to fetch a specific resource, such as an XMLHttpRequest, <SVG>, image, or script.

    References:
        https://developer.mozilla.org/en-US/docs/web/api/performanceresourcetiming
    """
    connect_end: float = Field(alias='connectEnd')
    connect_start: float = Field(alias='connectStart')
    decoded_body_size: int = Field(alias='decodedBodySize')
    domain_lookup_end: float = Field(alias='domainLookupEnd')
    domain_lookup_start: float = Field(alias='domainLookupStart')
    duration: float
    encoded_body_size: int = Field(alias='encodedBodySize')
    entry_type: str = Field(alias='entryType', default='resource')
    fetch_start: float = Field(alias='fetchStart')
    initiator_type: str = Field(alias='initiatorType')
    name: str
    next_hop_protocol: str = Field(alias='nextHopProtocol')
    redirect_end: float = Field(alias='redirectEnd')
    redirect_start: float = Field(alias='redirectStart')
    request_start: float = Field(alias='requestStart')
    response_end: float = Field(alias='responseEnd')
    response_start: float = Field(alias='responseStart')
    secure_connection_start: float = Field(alias='secureConnectionStart')
    server_timing: List = Field(alias='serverTiming')
    start_time: float = Field(alias='startTime')
    transfer_size: int = Field(alias='transferSize')
    worker_start: float = Field(alias='workerStart')


class WebPerformance(BaseModel):
    """ Pylenium's WebPerformance Object.

    This is built using multiple W3C Performance Timing objects to provide
    custom data points like:

    * Page Load Time
    * Time to First Contentful Paint
    * Time to Interactive (TTI)
    * and more!
    """
    time_origin: float  # High resolution timestamp of the start time of the Performance measurement
    navigation_timing: NavigationTiming
    paint_timing: PaintTiming
    resources: List[ResourceTiming]

    def page_load_time(self) -> float:
        """ The time it takes for the page to load as experienced by the user. """
        return self.navigation_timing.load_event_end - self.navigation_timing.start_time

    def time_to_first_byte(self) -> float:
        """ The time it takes before the first byte of response is received from the server. """
        return self.navigation_timing.response_start

    def time_to_first_contentful_paint(self) -> float:
        """ The time it takes for the majority of content to be fully rendered and consumable by the user. """
        return self.paint_timing.start_time

    def time_to_interactive(self) -> float:
        """ The time it takes for the layout to be stabilized and the page is responsive. """
        return self.navigation_timing.dom_complete

    def number_of_requests(self) -> int:
        """ The number of requests sent from start of navigation until end of page load. """
        return len(self.resources)

    def time_to_dom_content_loaded(self) -> float:
        return self.navigation_timing.dom_content_loaded_event_end

    def page_weight(self) -> float:
        """ The amount of bytes transferred for the page to be loaded. """
        resource_transfer_size = sum([r.transfer_size for r in self.resources])
        return self.navigation_timing.transfer_size + resource_transfer_size

    def connection_time(self) -> float:
        """ The time taken to connect to the server. """
        return self.navigation_timing.connect_end - self.navigation_timing.connect_start

    def request_time(self) -> float:
        """ The time taken to send a request to the server and receive the response. """
        return self.navigation_timing.response_end - self.navigation_timing.response_start

    def fetch_time(self) -> float:
        """ The time to complete the document fetch (including accessing any caches, etc.). """
        return self.navigation_timing.response_end - self.navigation_timing.fetch_start



from setuptools import setup, find_packages
import pylenium as app


setup(
    name='pyleniumio',
    version=app.__version__,
    packages=find_packages(),
    include_package_data=True,
    url='https://github.com/ElSnoMan/pyleniumio',
    license='MIT',
    author='Carlos Kidman',
    author_email='carlos@qap.dev',
    description='The best of Selenium and Cypress in a single Python Package',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    install_requires=[
        'selenium',
        'pytest',
        'pytest-xdist',
        'pytest-parallel',
        'pydantic',
        'pytest-reportportal',
        'Faker',
        'requests',
        'webdriver-manager',
        'click',
        'pyfiglet',
        'axe-selenium-python',
    ],
    data_files=[
        (
            '',
            [
                'pylenium/scripts/pylenium.json',
                'pylenium/scripts/pytest.ini',
                'pylenium/scripts/drag_and_drop.js',
                'pylenium/scripts/load_jquery.js',
            ],
        )
    ],
    entry_points='''
        [console_scripts]
        pylenium=pylenium.scripts.cli:cli
    ''',
)



from selenium.webdriver.support import expected_conditions as ec
from selenium.common.exceptions import NoSuchFrameException
from pylenium.element import Element


class FrameIsAvailable:
    """Expected Condition since the current one from Selenium doesn't work for strings, only tuples."""

    def __init__(self, frame_name_or_id):
        self.frame_name_or_id = frame_name_or_id

    def __call__(self, driver):
        try:
            driver.switch_to.frame(self.frame_name_or_id)
            return True
        except NoSuchFrameException:
            return False


class SwitchTo:
    def __init__(self, pylenium):
        self._py = pylenium

    def frame(self, name_or_id: str, timeout: int = 0):
        """Switch the driver's context to a frame given the name or id of the element.

        Args:
            name_or_id: The frame's `id` or `name` attribute value
            timeout: The number of seconds to wait for the frame to be switched to.

        Examples:
            # Switch to an iframe
            py.switch_to.frame('main-frame')
        """
        self._py.log.info(f"[STEP] py.switch_to.frame() - Switch to frame using name or id: ``{name_or_id}``")
        self._py.wait(timeout).until(FrameIsAvailable(name_or_id))
        return self._py

    def frame_by_element(self, element: Element, timeout: int = 0):
        """Switch the driver's context to the given frame element.

        Args:
            element (Element): The frame element to switch to
            timeout: The number of seconds to wait for the frame to be switched to.

        Examples:
            iframe = py.get('iframe')
            py.switch_to.frame_by_element(iframe)
        """
        self._py.log.info("[STEP] py.switch_to.frame_by_element() - Switch to frame using an Element.")
        self._py.wait(timeout).until(ec.frame_to_be_available_and_switch_to_it(element.locator))
        return self._py

    def parent_frame(self):
        """Switch the driver's context to the parent frame.

        If the parent frame is the current context, nothing happens.
        """
        self._py.log.info("[STEP] py.switch_to.parent_frame() - Switch to the parent frame")
        self._py.webdriver.switch_to.parent_frame()
        return self._py

    def default_content(self):
        """Switch the driver's context to the default content."""
        self._py.log.info("[STEP] py.switch_to.default_content() - Switch to default content of this browser session")
        self._py.webdriver.switch_to.default_content()
        return self._py

    def new_window(self):
        """Open a new Browser Window and switch the driver's context (aka focus) to it."""
        self._py.webdriver.switch_to.new_window("window")
        return self._py

    def new_tab(self):
        """Open a new Browser Tab and switch the driver's context (aka focus) to it."""
        self._py.webdriver.switch_to.new_window("tab")
        return self._py

    def window(self, name_or_handle="", index=0):
        """Switch the driver's context (aka focus) to the specified Browser Window or Browser Tab.

        Args:
            name_or_handle: The name or window handle of the Window or Tab to switch to.
            index: The index position of the Window Handle.

        * `index=0` would be the default content.

        Examples:
            # Switch to a Window by handle
            windows = py.window_handles
            py.switch_to.window(name_or_handle=windows[1])

            # Switch to a newly opened Browser Tab by index
            py.switch_to.window(index=1)
        """
        if index:
            handle = self._py.webdriver.window_handles[index]
            self._py.log.info(f"[STEP] py.switch_to.window() - Switch to a Tab or Window by index: ``{index}``")
            self._py.webdriver.switch_to.window(handle)
            return self._py
        elif name_or_handle:
            self._py.log.info(
                f"[STEP] py.switch_to.window() - Switch to Tab or Window by name or handle: ``{name_or_handle}``"
            )
            self._py.webdriver.switch_to.window(name_or_handle)
            return self._py
        else:
            # context unchanged
            return self._py
